"""
This module contains a wrapper for EPyT that allows a better error/warning handling.
"""
import warnings
from ctypes import byref, create_string_buffer
from epyt import epanet
from epyt.epanet import epanetapi, epanetmsxapi


class EPyT(epanet):
    """
    Wrapper for the `epyt.epanet <https://epanet-python-toolkit-epyt.readthedocs.io/en/latest/api.html#epyt.epanet.epanet>`_ class.
    This wrapper adds functionalities for getting error/warning messages after each
    EPANET function call.
    """
    def __init__(self, *argv, version=2.2, ph=False, loadfile=False, customlib=None,
                 display_msg=True, display_warnings=True):
        super().__init__(*argv, version=version, ph=ph, loadfile=loadfile, customlib=customlib,
                         display_msg=display_msg, display_warnings=display_warnings)

        # Inject custom EPANET API wrapper
        if self.api._ph is not None:
            self.api.ENdeleteproject()
        else:
            self.api.ENclose()

        self.api = MyEpanetAPI(version=version, ph=ph, customlib=customlib)

        self.api.ENopen(self.TempInpFile, self.RptTempfile, self.BinTempfile)

        self.LibEPANETpath = self.api.LibEPANETpath
        self.LibEPANET = self.api.LibEPANET

    def _logFunctionError(self, function_name):
        # Do not print warnings.
        pass

    def loadMSXFile(self, msxname, customMSXlib=None, ignore_properties=False):
        super().loadMSXFile(msxname, customMSXlib, ignore_properties)

        # Inject custom EPANET-MSX API wrapper
        self.msx.MSXclose()
        self.msx = MyEpanetMsxAPI(msxfile=self.MSXTempFile, customMSXlib=customMSXlib,
                                  display_msg=self.display_msg,
                                  msxrealfile=self.MSXFile)
        return self.msx

    def set_error_handling(self, raise_exception_on_error: bool, warn_on_error: bool,
                           ignore_error_codes: list[int]) -> None:
        """
        Specifies the behavior in the case of an error/warning --
        i.e. should an exception or warning be raised or not?

        Parameters
        ----------
        raise_exception_on_error : `bool`
            True if an exception should be raise, False otherwise.
        warn_on_error : `bool`
            True if a warning should be generated, False otherwise.
        ignore_error_codes : `list[int]`
            List of error codes that should be ignored -- i.e., no exception or
            warning will be generated.
        """
        self.api.set_error_handling(raise_exception_on_error, warn_on_error, ignore_error_codes)
        if self.msx is not None:
            self.msx.set_error_handling(raise_exception_on_error, warn_on_error, ignore_error_codes)

    def was_last_func_successful(self) -> bool:
        """
        Checks if the last EPANET function call was successful.

        Returns
        -------
        `bool`
            True if there was no error/warning, False otherwise.
        """
        return self.api.was_last_func_successful()

    def get_last_error_desc(self) -> str:
        """
        Returns a description of the last error/warning if any occured.
        Returns None if there was no error/warning.

        Returns
        -------
        `str`
            Error/warning description.
        """
        return self.api.get_last_error_desc()

    def get_last_error_code(self) -> int:
        """
        Returns the code of the last error/warning if any occured.
        Returns 0 if there was no error/warning.

        Refer to the `EPANET documentation <http://wateranalytics.org/EPANET/group___warning_codes.html>`_
        for a list of all possible warning codes and their meanings.

        Returns
        -------
        `int`
            Error/warning code.
        """
        return self.api.get_last_error_code()


class MyEpanetMsxAPI(epanetmsxapi):
    """
    Wrapper for the `epyt.epanet.epanetmsxapi <https://epanet-python-toolkit-epyt.readthedocs.io/en/latest/api.html#epyt.epanet.epanetmsxapi>`_
    class adding error/warning storage functionalities.
    """
    def __init__(self, raise_on_error: bool = True, warn_on_error: bool = False,
                 ignore_error_codes: list[int] = [], **kwds):
        self.__raise_on_error = raise_on_error
        self.__warn_on_error = warn_on_error
        self.__ignore_error_codes = ignore_error_codes
        self.__last_error_code = None
        self.__last_error_desc = None

        super().__init__(**kwds)

    def _logFunctionError(self, function_name):
        # Do not print warnings.
        pass

    def set_error_handling(self, raise_on_error: bool, warn_on_error: bool,
                           ignore_error_codes: list[int] = []) -> None:
        """
        Specifies the behavior in the case of an error/warning --
        i.e. should an exception or warning be raised or not?

        Parameters
        ----------
        raise_exception_on_error : `bool`
            True if an exception should be raise, False otherwise.
        warn_on_error : `bool`
            True if a warning should be generated, False otherwise.
        ignore_error_codes : `list[int]`
            List of error codes that should be ignored -- i.e., no exception or
            warning will be generated.
        """
        self.__raise_on_error = raise_on_error
        self.__warn_on_error = warn_on_error
        self.__ignore_error_codes = ignore_error_codes

    def get_last_error_desc(self) -> str:
        """
        Returns the description of the last EPANET-MSX error/warning (if any).

        Returns
        -------
        `str`
            Description of the last error/warning. None, if there was no error/warning.
        """
        return self.__last_error_desc

    def get_last_error_code(self) -> int:
        """
        Returns the code of the last EPANET-MSX error/warnning (if any).

        Refer to the EPANET-MSX user manual for a list of all possible warning codes
        and their meanings.

        Returns
        -------
        `int`
            Code of the last error/warning. 0, if there was no error/warning.
        """
        return self.__last_error_code

    def was_last_func_successful(self) -> bool:
        """
        Checks if the last EPANET-MSX call was successful or not.

        Parameters
        ----------
        `bool`
            True if the last EPANET-MSX call returned an error/warning, False otherwise.
        """
        return self.__last_error_desc is None

    def _reset_error(self) -> None:
        self.__last_error_code = 0
        self.__last_error_desc = None

    def MSXerror(self, err_code: int) -> None:
        error_desc = create_string_buffer(256)
        self.msx_error(err_code, error_desc, 256)
        self.__last_error_code = err_code
        self.__last_error_desc = error_desc.value.decode()

        if self.__last_error_code not in self.__ignore_error_codes:
            if self.__warn_on_error:
                warnings.warn(self.__last_error_desc, RuntimeWarning)
            if self.__raise_on_error:
                raise RuntimeError(self.__last_error_desc)

    def MSXopen(self, msxfile, msxrealfile):
        self._reset_error()
        super().MSXopen(msxfile, msxrealfile)

    def MSXclose(self):
        self._reset_error()
        return super().MSXclose()

    def MSXgetindex(self, obj_type, obj_id):
        self._reset_error()
        return super().MSXgetindex(obj_type, obj_id)

    def MSXgetID(self, obj_type, index, id_len=80):
        self._reset_error()
        return super().MSXgetID(obj_type, index, id_len)

    def MSXgetIDlen(self, obj_type, index):
        self._reset_error()
        return super().MSXgetIDlen(obj_type, index)

    def MSXgetspecies(self, index):
        self._reset_error()
        return super().MSXgetspecies(index)

    def MSXgetcount(self, code):
        self._reset_error()
        return super().MSXgetcount(code)

    def MSXgetconstant(self, index):
        self._reset_error()
        return super().MSXgetconstant(index)

    def MSXgetparameter(self, obj_type, index, param):
        self._reset_error()
        return super().MSXgetparameter(obj_type, index, param)

    def MSXgetpatternlen(self, pattern_index):
        self._reset_error()
        return super().MSXgetpatternlen(pattern_index)

    def MSXgetpatternvalue(self, pattern_index, period):
        self._reset_error()
        return super().MSXgetpatternvalue(pattern_index, period)

    def MSXgetinitqual(self, obj_type, index, species):
        self._reset_error()
        return super().MSXgetinitqual(obj_type, index, species)

    def MSXgetsource(self, node_index, species_index):
        self._reset_error()
        return super().MSXgetsource(node_index, species_index)

    def MSXsaveoutfile(self, filename):
        self._reset_error()
        super().MSXsaveoutfile(filename)

    def MSXsavemsxfile(self, filename):
        self._reset_error()
        super().MSXsavemsxfile(filename)

    def MSXsetconstant(self, index, value):
        self._reset_error()
        super().MSXsetconstant(index, value)

    def MSXsetparameter(self, obj_type, index, param, value):
        self._reset_error()
        super().MSXsetparameter(obj_type, index, param, value)

    def MSXsetinitqual(self, obj_type, index, species, value):
        self._reset_error()
        super().MSXsetinitqual(obj_type, index, species, value)

    def MSXsetpattern(self, index, factors, nfactors):
        self._reset_error()
        super().MSXsetpattern(index, factors, nfactors)

    def MSXsetpatternvalue(self, pattern, period, value):
        self._reset_error()
        super().MSXsetpatternvalue(pattern, period, value)

    def MSXsolveQ(self):
        self._reset_error()
        super().MSXsolveQ()

    def MSXsolveH(self):
        self._reset_error()
        super().MSXsolveH()

    def MSXaddpattern(self, pattern_id):
        self._reset_error()
        super().MSXaddpattern(pattern_id)

    def MSXusehydfile(self, filename):
        self._reset_error()
        super().MSXusehydfile(filename)

    def MSXstep(self):
        self._reset_error()
        return super().MSXstep()

    def MSXinit(self, flag):
        self._reset_error()
        super().MSXinit(flag)

    def MSXreport(self):
        self._reset_error()
        super().MSXreport()

    def MSXgetqual(self, type, index, species):
        self._reset_error()
        return super().MSXgetqual(type, index, species)

    def MSXsetsource(self, node, species, type, level, pat):
        self._reset_error()
        super().MSXsetsource(node, species, type, level, pat)


class MyEpanetAPI(epanetapi):
    """
    Wrapper for the `epyt.epanet.epanetapi <https://epanet-python-toolkit-epyt.readthedocs.io/en/latest/api.html#epyt.epanet.epanetapi>`_
    class adding error/warning storage functionalities.
    """
    def __init__(self, raise_on_error: bool = True, warn_on_error: bool = False,
                 ignore_error_codes: list[int] = [], **kwds):
        self.__raise_on_error = raise_on_error
        self.__warn_on_error = warn_on_error
        self.__ignore_error_codes = ignore_error_codes
        self.__last_error_code = None
        self.__last_error_desc = None

        super().__init__(**kwds)

    def set_error_handling(self, raise_on_error: bool, warn_on_error: bool,
                           ignore_error_codes: list[int] = []) -> None:
        """
        Specifies the behavior in the case of an error/warning --
        i.e. should an exception or warning be raised or not?

        Parameters
        ----------
        raise_exception_on_error : `bool`
            True if an exception should be raise, False otherwise.
        warn_on_error : `bool`
            True if a warning should be generated, False otherwise.
        ignore_error_codes : `list[int]`
            List of error codes that should be ignored -- i.e., no exception or
            warning will be generated.
        """
        self.__raise_on_error = raise_on_error
        self.__warn_on_error = warn_on_error
        self.__ignore_error_codes = ignore_error_codes

    def get_last_error_desc(self) -> str:
        """
        Returns the description of the last EPANET error/warning (if any).

        Returns
        -------
        `str`
            Description of the last error/warning. None, if there was no error/warning.
        """
        return self.__last_error_desc

    def get_last_error_code(self) -> int:
        """
        Returns the code of the last EPANET error/warnning (if any).

        Refer to the `EPANET documentation <http://wateranalytics.org/EPANET/group___warning_codes.html>`_
        for a list of all possible warning codes and their meanings.

        Returns
        -------
        `int`
            Code of the last error/warning. 0, if there was no error/warning.
        """
        return self.__last_error_code

    def was_last_func_successful(self) -> bool:
        """
        Checks if the last EPANET call was successful or not.

        Parameters
        ----------
        `bool`
            True if the last EPANET call returned an error/warning, False otherwise.
        """
        return self.__last_error_desc is None

    def _reset_error(self) -> None:
        self.__last_error_code = 0
        self.__last_error_desc = None

    def _check_for_error(self) -> None:
        if self.errcode != 0:
            error_desc = create_string_buffer(150)
            self._lib.ENgeterror(self.errcode, byref(error_desc), 150)

            self.__last_error_code = self.errcode
            self.__last_error_desc = error_desc.value.decode()

            if self.__last_error_code not in self.__ignore_error_codes:
                if self.__warn_on_error:
                    warnings.warn(self.__last_error_desc, RuntimeWarning)
                if self.__raise_on_error:
                    raise RuntimeError(self.__last_error_desc)

    def ENepanet(self, inpfile="", rptfile="", binfile=""):
        self._reset_error()
        super().ENepanet(inpfile, rptfile, binfile)
        self._check_for_error()

    def ENaddcontrol(self, conttype, lindex, setting, nindex, level):
        self._reset_error()
        r = super().ENaddcontrol(conttype, lindex, setting, nindex, level)
        self._check_for_error()
        return r

    def ENaddcurve(self, cid):
        self._reset_error()
        super().ENaddcurve(cid)
        self._check_for_error()

    def ENadddemand(self, nodeIndex, baseDemand, demandPattern, demandName):
        self._reset_error()
        super().ENadddemand(nodeIndex, baseDemand, demandPattern, demandName)
        self._check_for_error()

    def ENaddlink(self, linkid, linktype, fromnode, tonode):
        self._reset_error()
        r = super().ENaddlink(linkid, linktype, fromnode, tonode)
        self._check_for_error()
        return r

    def ENaddnode(self, nodeid, nodetype):
        self._reset_error()
        r = super().ENaddnode(nodeid, nodetype)
        self._check_for_error()
        return r

    def ENaddpattern(self, patid):
        self._reset_error()
        super().ENaddpattern(patid)
        self._check_for_error()

    def ENaddrule(self, rule):
        self._reset_error()
        super().ENaddrule(rule)
        self._check_for_error()

    def ENclearreport(self):
        self._reset_error()
        super().ENclearreport()
        self._check_for_error()

    def ENclose(self):
        self._reset_error()
        super().ENclose()
        self._check_for_error()

    def ENcloseH(self):
        self._reset_error()
        super().ENcloseH()
        self._check_for_error()

    def ENcloseQ(self):
        self._reset_error()
        super().ENcloseQ()
        self._check_for_error()

    def ENcopyreport(self, filename):
        self._reset_error()
        super().ENcopyreport(filename)
        self._check_for_error()

    def ENcreateproject(self):#
        self._reset_error()
        super().ENcreateproject()
        self._check_for_error()

    def ENdeletecontrol(self, index):
        self._reset_error()
        super().ENdeletecontrol(index)
        self._check_for_error()

    def ENdeletecurve(self, indexCurve):
        self._reset_error()
        super().ENdeletecurve(indexCurve)
        self._check_for_error()

    def ENdeletedemand(self, nodeIndex, demandIndex):
        self._reset_error()
        super().ENdeletedemand(nodeIndex, demandIndex)
        self._check_for_error()

    def ENdeletelink(self, indexLink, condition):
        self._reset_error()
        super().ENdeletelink(indexLink, condition)
        self._check_for_error()

    def ENdeletenode(self, indexNode, condition):
        self._reset_error()
        super().ENdeletenode(indexNode, condition)
        self._check_for_error()

    def ENdeletepattern(self, indexPat):
        self._reset_error()
        super().ENdeletepattern(indexPat)
        self._check_for_error()

    def ENdeleteproject(self):
        self._reset_error()
        super().ENdeleteproject()
        self._check_for_error()

    def ENdeleterule(self, index):
        self._reset_error()
        super().ENdeleterule(index)
        self._check_for_error()

    def ENgetaveragepatternvalue(self, index):
        self._reset_error()
        r = super().ENgetaveragepatternvalue(index)
        self._check_for_error()
        return r

    def ENgetbasedemand(self, index, numdemands):
        self._reset_error()
        r = super().ENgetbasedemand(index, numdemands)
        self._check_for_error()
        return r

    def ENgetcomment(self, object_, index):
        self._reset_error()
        r = super().ENgetcomment(object_, index)
        self._check_for_error()
        return r

    def ENgetcontrol(self, cindex):
        self._reset_error()
        r = super().ENgetcontrol(cindex)
        self._check_for_error()
        return r

    def ENgetcoord(self, index):
        self._reset_error()
        r = super().ENgetcoord(index)
        self._check_for_error()
        return r

    def ENgetcount(self, countcode):
        self._reset_error()
        r = super().ENgetcount(countcode)
        self._check_for_error()
        return r

    def ENgetcurve(self, index):
        self._reset_error()
        r = super().ENgetcurve(index)
        self._check_for_error()
        return r

    def ENgetcurveid(self, index):
        self._reset_error()
        r = super().ENgetcurveid(index)
        self._check_for_error()
        return r

    def ENgetcurveindex(self, Id):
        self._reset_error()
        r = super().ENgetcurveindex(Id)
        self._check_for_error()
        return r

    def ENgetcurvelen(self, index):
        self._reset_error()
        r = super().ENgetcurvelen(index)
        self._check_for_error()
        return r

    def ENgetcurvetype(self, index):
        self._reset_error()
        r = super().ENgetcurvetype(index)
        self._check_for_error()
        return r

    def ENgetcurvevalue(self, index, period):
        self._reset_error()
        r = super().ENgetcurvevalue(index, period)
        self._check_for_error()
        return r

    def ENgetdemandindex(self, nodeindex, demandName):
        self._reset_error()
        r = super().ENgetdemandindex(nodeindex, demandName)
        self._check_for_error()
        return r

    def ENgetdemandmodel(self):
        self._reset_error()
        r = super().ENgetdemandmodel()
        self._check_for_error()
        return r

    def ENgetdemandname(self, node_index, demand_index):
        self._reset_error()
        r = super().ENgetdemandname(node_index, demand_index)
        self._check_for_error()
        return r

    def ENgetdemandpattern(self, index, numdemands):
        self._reset_error()
        r = super().ENgetdemandpattern(index,numdemands)
        self._check_for_error()
        return r

    def ENgetelseaction(self, ruleIndex, actionIndex):
        self._reset_error()
        r = super().ENgetelseaction(ruleIndex, actionIndex)
        self._check_for_error()
        return r

    def ENgetflowunits(self):
        self._reset_error()
        r = super().ENgetflowunits()
        self._check_for_error()
        return r

    def ENgetheadcurveindex(self, pumpindex):
        self._reset_error()
        r = super().ENgetheadcurveindex(pumpindex)
        self._check_for_error()
        return r

    def ENgetlinkid(self, index):
        self._reset_error()
        r = super().ENgetlinkid(index)
        self._check_for_error()
        return r

    def ENgetlinkindex(self, Id):
        self._reset_error()
        r = super().ENgetlinkindex(Id)
        self._check_for_error()
        return r

    def ENgetlinknodes(self, index):
        self._reset_error()
        r = super().ENgetlinknodes(index)
        self._check_for_error()
        return r

    def ENgetlinktype(self, index):
        self._reset_error()
        r = super().ENgetlinktype(index)
        self._check_for_error()
        return r

    def ENgetlinkvalue(self, index, paramcode):
        self._reset_error()
        r = super().ENgetlinkvalue(index, paramcode)
        self._check_for_error()
        return r

    def ENgetnodeid(self, index):
        self._reset_error()
        r = super().ENgetnodeid(index)
        self._check_for_error()
        return r

    def ENgetnodeindex(self, Id):
        self._reset_error()
        r = super().ENgetnodeindex(Id)
        self._check_for_error()
        return r

    def ENgetnodetype(self, index):
        self._reset_error()
        r = super().ENgetnodetype(index)
        self._check_for_error()
        return r

    def ENgetnodevalue(self, index, code_p):
        self._reset_error()
        r = super().ENgetnodevalue(index, code_p)
        self._check_for_error()
        return r

    def ENgetnumdemands(self, index):
        self._reset_error()
        r = super().ENgetnumdemands(index)
        self._check_for_error()
        return r

    def ENgetoption(self, optioncode):
        self._reset_error()
        r = super().ENgetoption(optioncode)
        self._check_for_error()
        return r

    def ENgetpatternid(self, index):
        self._reset_error()
        r = super().ENgetpatternid(index)
        self._check_for_error()
        return r

    def ENgetpatternindex(self, Id):
        self._reset_error()
        r = super().ENgetpatternindex(Id)
        self._check_for_error()
        return r

    def ENgetpatternlen(self, index):
        self._reset_error()
        r = super().ENgetpatternlen(index)
        self._check_for_error()
        return r

    def ENgetpatternvalue(self, index, period):
        self._reset_error()
        r = super().ENgetpatternvalue(index, period)
        self._check_for_error()
        return r

    def ENgetpremise(self, ruleIndex, premiseIndex):
        self._reset_error()
        r = super().ENgetpremise(ruleIndex, premiseIndex)
        self._check_for_error()
        return r

    def ENgetpumptype(self, index):
        self._reset_error()
        r = super().ENgetpumptype(index)
        self._check_for_error()
        return r

    def ENgetqualinfo(self):
        self._reset_error()
        r = super().ENgetqualinfo()
        self._check_for_error()
        return r

    def ENgetqualtype(self):
        self._reset_error()
        r = super().ENgetqualtype()
        self._check_for_error()
        return r

    def ENgetresultindex(self, objecttype, index):
        self._reset_error()
        r = super().ENgetresultindex(objecttype, index)
        self._check_for_error()
        return r

    def ENgetrule(self, index):
        self._reset_error()
        r = super().ENgetrule(index)
        self._check_for_error()
        return r

    def ENgetruleID(self, index):
        self._reset_error()
        r = super().ENgetruleID(index)
        self._check_for_error()
        return r

    def ENgetstatistic(self, code):
        self._reset_error()
        r = super().ENgetstatistic(code)
        self._check_for_error()
        return r

    def ENgetthenaction(self, ruleIndex, actionIndex):
        self._reset_error()
        r = super().ENgetthenaction(ruleIndex, actionIndex)
        self._check_for_error()
        return r

    def ENgettimeparam(self, paramcode):
        self._reset_error()
        r = super().ENgettimeparam(paramcode)
        self._check_for_error()
        return r

    def ENgettitle(self):
        self._reset_error()
        r = super().ENgettitle()
        self._check_for_error()
        return r

    def ENgetversion(self):
        self._reset_error()
        r = super().ENgetversion()
        self._check_for_error()
        return r

    def ENgetvertex(self, index, vertex):
        self._reset_error()
        r = super().ENgetvertex(index, vertex)
        self._check_for_error()
        return r

    def ENgetvertexcount(self, index):
        self._reset_error()
        r = super().ENgetvertexcount(index)
        self._check_for_error()
        return r

    def ENinit(self, unitsType, headLossType):
        self._reset_error()
        super().ENinit(unitsType, headLossType)
        self._check_for_error()

    def ENinitH(self, flag):
        self._reset_error()
        super().ENinitH(flag)
        self._check_for_error()

    def ENinitQ(self, saveflag):
        self._reset_error()
        super().ENinitQ(saveflag)
        self._check_for_error()

    def ENnextH(self):
        self._reset_error()
        r = super().ENnextH()
        self._check_for_error()
        return r

    def ENnextQ(self):
        self._reset_error()
        r = super().ENnextQ()
        self._check_for_error()
        return r

    def ENopen(self, inpname=None, repname=None, binname=None):
        self._reset_error()
        super().ENopen(inpname, repname, binname)
        self._check_for_error()

    def ENopenH(self):
        self._reset_error()
        super().ENopenH()
        self._check_for_error()

    def ENopenQ(self):
        self._reset_error()
        super().ENopenQ()
        self._check_for_error()

    def ENreport(self):
        self._reset_error()
        super().ENreport()
        self._check_for_error()

    def ENresetreport(self):
        self._reset_error()
        super().ENresetreport()
        self._check_for_error()

    def ENrunH(self):
        self._reset_error()
        r = super().ENrunH()
        self._check_for_error()
        return r

    def ENrunQ(self):
        self._reset_error()
        r = super().ENrunQ()
        self._check_for_error()
        return r

    def ENsaveH(self):
        self._reset_error()
        super().ENsaveH()
        self._check_for_error()

    def ENsavehydfile(self, fname):
        self._reset_error()
        super().ENsavehydfile(fname)
        self._check_for_error()

    def ENsaveinpfile(self, inpname):
        self._reset_error()
        super().ENsaveinpfile(inpname)
        self._check_for_error()

    def ENsetbasedemand(self, index, demandIdx, value):
        self._reset_error()
        super().ENsetbasedemand(index, demandIdx, value)
        self._check_for_error()

    def ENsetcomment(self, object_, index, comment):
        self._reset_error()
        super().ENsetcomment(object_, index, comment)
        self._check_for_error()

    def ENsetcontrol(self, cindex, ctype, lindex, setting, nindex, level):
        self._reset_error()
        super().ENsetcontrol(cindex, ctype, lindex, setting, nindex, level)
        self._check_for_error()

    def ENsetcoord(self, index, x, y):
        self._reset_error()
        super().ENsetcoord(index, x, y)
        self._check_for_error()

    def ENsetcurve(self, index, x, y, nfactors):
        self._reset_error()
        super().ENsetcurve(index, x, y, nfactors)
        self._check_for_error()

    def ENsetcurveid(self, index, Id):
        self._reset_error()
        super().ENsetcurveid(index, Id)
        self._check_for_error()

    def ENsetcurvevalue(self, index, pnt, x, y):
        self._reset_error()
        super().ENsetcurvevalue(index, pnt, x, y)
        self._check_for_error()

    def ENsetdemandmodel(self, Type, pmin, preq, pexp):
        self._reset_error()
        super().ENsetdemandmodel(Type, pmin, preq, pexp)
        self._check_for_error()

    def ENsetdemandname(self, node_index, demand_index, demand_name):
        self._reset_error()
        super().ENsetdemandname(node_index, demand_index, demand_name)
        self._check_for_error()

    def ENsetdemandpattern(self, index, demandIdx, patInd):
        self._reset_error()
        super().ENsetdemandpattern(index, demandIdx, patInd)
        self._check_for_error()

    def ENsetelseaction(self, ruleIndex, actionIndex, linkIndex, status, setting):
        self._reset_error()
        super().ENsetelseaction(ruleIndex, actionIndex, linkIndex, status, setting)
        self._check_for_error()

    def ENsetflowunits(self, code):
        self._reset_error()
        super().ENsetflowunits(code)
        self._check_for_error()

    def ENsetheadcurveindex(self, pumpindex, curveindex):
        self._reset_error()
        super().ENsetheadcurveindex(pumpindex, curveindex)
        self._check_for_error()

    def ENsetjuncdata(self, index, elev, dmnd, dmndpat):
        self._reset_error()
        super().ENsetjuncdata(index, elev, dmnd, dmndpat)
        self._check_for_error()

    def ENsetlinkid(self, index, newid):
        self._reset_error()
        super().ENsetlinkid(index, newid)
        self._check_for_error()

    def ENsetlinknodes(self, index, startnode, endnode):
        self._reset_error()
        super().ENsetlinknodes(index, startnode, endnode)
        self._check_for_error()

    def ENsetlinktype(self, indexLink, paramcode, actionCode):
        self._reset_error()
        r = super().ENsetlinktype(indexLink, paramcode, actionCode)
        self._check_for_error()
        return r

    def ENsetlinkvalue(self, index, paramcode, value):
        self._reset_error()
        super().ENsetlinkvalue(index, paramcode, value)
        self._check_for_error()

    def ENsetnodeid(self, index, newid):
        self._reset_error()
        super().ENsetnodeid(index, newid)
        self._check_for_error()

    def ENsetnodevalue(self, index, paramcode, value):
        self._reset_error()
        super().ENsetnodevalue(index, paramcode, value)
        self._check_for_error()

    def ENsetoption(self, optioncode, value):
        self._reset_error()
        super().ENsetoption(optioncode, value)
        self._check_for_error()

    def ENsetpattern(self, index, factors, nfactors):
        self._reset_error()
        super().ENsetpattern(index, factors, nfactors)
        self._check_for_error()

    def ENsetpatternid(self, index, Id):
        self._reset_error()
        super().ENsetpatternid(index, Id)
        self._check_for_error()

    def ENsetpatternvalue(self, index, period, value):
        self._reset_error()
        super().ENsetpatternvalue(index, period, value)
        self._check_for_error()

    def ENsetpipedata(self, index, length, diam, rough, mloss):
        self._reset_error()
        super().ENsetpipedata(index, length, diam, rough, mloss)
        self._check_for_error()

    def ENsetpremise(self, ruleIndex, premiseIndex, logop, object_, objIndex, variable,
                     relop, status, value):
        self._reset_error()
        super().ENsetpremise(ruleIndex, premiseIndex, logop, object_, objIndex, variable,
                             relop, status, value)
        self._check_for_error()

    def ENsetpremiseindex(self, ruleIndex, premiseIndex, objIndex):
        self._reset_error()
        super().ENsetpremiseindex(ruleIndex, premiseIndex, objIndex)
        self._check_for_error()

    def ENsetpremisestatus(self, ruleIndex, premiseIndex, status):
        self._reset_error()
        super().ENsetpremisestatus(ruleIndex, premiseIndex, status)
        self._check_for_error()

    def ENsetpremisevalue(self, ruleIndex, premiseIndex, value):
        self._reset_error()
        super().ENsetpremisevalue(ruleIndex, premiseIndex, value)
        self._check_for_error()

    def ENsetqualtype(self, qualcode, chemname, chemunits, tracenode):
        self._reset_error()
        super().ENsetqualtype(qualcode, chemname, chemunits, tracenode)
        self._check_for_error()

    def ENsetreport(self, command):
        self._reset_error()
        super().ENsetreport(command)
        self._check_for_error()

    def ENsetrulepriority(self, ruleIndex, priority):
        self._reset_error()
        super().ENsetrulepriority(ruleIndex, priority)
        self._check_for_error()

    def ENsetstatusreport(self, statuslevel):
        self._reset_error()
        super().ENsetstatusreport(statuslevel)
        self._check_for_error()

    def ENsettankdata(self, index, elev, initlvl, minlvl, maxlvl, diam, minvol, volcurve):
        self._reset_error()
        super().ENsettankdata(index, elev, initlvl, minlvl, maxlvl, diam, minvol, volcurve)
        self._check_for_error()

    def ENsetthenaction(self, ruleIndex, actionIndex, linkIndex, status, setting):
        self._reset_error()
        super().ENsetthenaction(ruleIndex, actionIndex, linkIndex, status, setting)
        self._check_for_error()

    def ENsettimeparam(self, paramcode, timevalue):
        self._reset_error()
        super().ENsettimeparam(paramcode, timevalue)
        self._check_for_error()

    def ENsettitle(self, line1, line2, line3):
        self._reset_error()
        super().ENsettitle(line1, line2, line3)
        self._check_for_error()

    def ENsetvertices(self, index, x, y, vertex):
        self._reset_error()
        super().ENsetvertices(index, x, y, vertex)
        self._check_for_error()

    def ENsolveH(self):
        self._reset_error()
        super().ENsolveH()
        self._check_for_error()

    def ENsolveQ(self):
        self._reset_error()
        super().ENsolveQ()
        self._check_for_error()

    def ENstepQ(self):
        self._reset_error()
        r = super().ENstepQ()
        self._check_for_error()
        return r

    def ENusehydfile(self, hydfname):
        self._reset_error()
        super().ENusehydfile(hydfname)
        self._check_for_error()

    def ENwriteline(self, line):
        self._reset_error()
        super().ENwriteline(line)
        self._check_for_error()



import subprocess
import warnings
import shutil
import sys
import os

with open(os.path.join(os.path.dirname(__file__), 'VERSION'), encoding="utf-8") as f:
    VERSION = f.read().strip()
    __version__ = VERSION


def compile_libraries_unix(lib_epanet_name: str, compile_script_name: str,
                           gcc_name: str = "gcc") -> None:
    """Compile EPANET and EPANET-MSX libraries if needed."""
    path_to_custom_libs = os.path.join(os.path.dirname(__file__), "customlibs")
    path_to_lib_epanet = os.path.join(path_to_custom_libs, lib_epanet_name)
    path_to_epanet = os.path.join(os.path.dirname(__file__), "EPANET")

    update = False
    if os.path.isfile(path_to_lib_epanet):
        if os.path.getmtime(__file__) > os.path.getmtime(path_to_lib_epanet):
            update = True

    if not os.path.isfile(path_to_lib_epanet) or update:
        if shutil.which(gcc_name) is not None:
            print("Compiling EPANET and EPANET-MSX...")
            try:
                subprocess.check_call(f"cd \"{path_to_epanet}\"; bash {compile_script_name}",
                                      shell=True)
                print("Done")
            except subprocess.CalledProcessError as ex:
                print(f"Compilation failed\n{ex}")
                warnings.warn("Falling back to pre-compiled library shipped by EPyT.")
        else:
            warnings.warn("GCC is not available to compile the required libraries.\n" +
                          "Falling back to pre-compiled library shipped by EPyT.")


if sys.platform.startswith("linux"):
    compile_libraries_unix("libepanet2_2.so", "compile_linux.sh")
elif sys.platform.startswith("darwin"):
    compile_libraries_unix("libepanet2_2.dylib", "compile_macos.sh", gcc_name="gcc-15")



"""
Module provides a function for loading the water usage data set by P. Pavlou et al.
"""
import os
import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, precision_score, roc_auc_score, f1_score

from ...utils import get_temp_folder, download_if_necessary


def compute_evaluation_score(y_pred: np.ndarray, y: np.ndarray) -> dict:
    """
    Evaluates the performance of a detection method.

    Note that instead of a single metric, the following set of metrics is used:
        - Accuracy
        - Precision
        - F1-score (using "micro" averaging)
        - Cohen's kappa
        - ROC AUC

    Parameters
    ----------
    y_pred : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Event indication prediction over time
    y : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Ground truth event indication over time.

    Returns
    -------
    `dict`
        All evaluation scores.
    """
    return {"accuracy": accuracy_score(y, y_pred),
            "precision": precision_score(y, y_pred, average="weighted"),
            "f1-micro": f1_score(y, y_pred, average="micro"), "roc-auc": roc_auc_score(y, y_pred)}


def load_water_usage(download_dir: str = None, return_X_y: bool = True, verbose: bool = True) -> dict:
    """
    "Monitoring domestic water consumption: A comparative study of model-based and data-driven
    end-use disaggregation methods" by P. Pavlou, S. Filippou, S. Solonos, S. G. Vrachimis,
    K. Malialis, D. G. Eliades, T. Theocharides, M. M. Polycarpou is a benchmark concerning the
    monitoring of water usage of different household appliances. Informing consumers about it has
    been shown to have an impact on their behavior toward drinking water conservation. The data
    were created using the STochastic Residential water End-use Model (STREaM)
    (Cominola et al., 2018), a modelling software developed that generates synthetic time series
    data of a household.

    This benchmark data set is for identifying active appliances from the aggregated water
    consumption -- i.e. a multi-class classification probelm. The data set considers the use
    of standard toilet, standard shower, standard faucet, high efficiency clothes washer,
    and standard dishwasher in a 2-person household for a period of 180 days (6 months) and
    it has a resolution of 10s.
    The data set is already split into 3 sub-sets for training (90 days), validation (45 days),
    and testing (45 days).

    For more information see https://github.com/KIOS-Research/Water-Usage-Dataset/

    .. note::

        Note that although this data set is synthetic, only the final data set is provided.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the data files -- if None, the temp folder will be used.
        If the path does not exist, the data files will be downloaded to the given path.

        The default is None.
    return_X_y : `bool`, optional
        If True, the data is returned together with the multi-class labels as two Numpy arrays,
        otherwise, the data is returned as Pandas data frame.

        The default is True.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.

    Returns
    -------
    `dict`
        The data set as a dictionary with entries "train", "validation", and "test" containing
        the respective data.
    """
    # Download data if necessary
    download_dir = download_dir if download_dir is not None else get_temp_folder()

    base_url = "https://github.com/KIOS-Research/Water-Usage-Dataset/raw/main/Dataset/"
    url_train_data = base_url + "Trainset.csv"
    url_valid_data = base_url + "Validationset.csv"
    url_test_data = base_url + "Testset.csv"

    f_train_in = os.path.join(download_dir, "train_water_usage.csv")
    f_valid_in = os.path.join(download_dir, "valid_water_usage.csv")
    f_test_in = os.path.join(download_dir, "test_water_usage.csv")

    download_if_necessary(f_train_in, url_train_data, verbose)
    download_if_necessary(f_valid_in, url_valid_data, verbose)
    download_if_necessary(f_test_in, url_test_data, verbose)

    # Load and return data
    df_data_train = pd.read_csv(f_train_in)
    df_data_valid = pd.read_csv(f_valid_in)
    df_data_test = pd.read_csv(f_test_in)

    if return_X_y is False:
        return {"train": df_data_train, "validation": df_data_valid, "test": df_data_test}
    else:
        r = {"train": None, "validation": None, "test": None}

        for k, df_data in zip(["train", "validation", "test"],
                              [df_data_train, df_data_valid, df_data_test]):
            X = df_data["TOTAL"].to_numpy()
            del df_data["TOTAL"]

            y = df_data.to_numpy()
            y = (y != 0).astype(np.int8)

            r[k] = (X, y)

        return r



"""
Module provides implementations of different types of actuator events.
"""
import warnings
from epyt.epanet import epanet
import numpy as np

from .system_event import SystemEvent
from ...serialization import serializable, JsonSerializable, PUMP_STATE_EVENT_ID, \
    PUMP_SPEED_EVENT_ID, VALVE_STATE_EVENT_ID


class ActuatorConstants:
    """
    Class defining some constants related to actuator events.

    Attributes
    ----------
    EN_CLOSED
        Valve or pump is closed.
    EN_OPEN
        Valve or pump is open -- i.e. active.
    """
    EN_CLOSED       = 0
    EN_OPEN         = 1


class ActuatorEvent(SystemEvent):
    """
    Base class of an actuator event.

    .. note::
        Note that actuator events are one-time events -- i.e.
        they are executed only once at a given point in time.

    Parameters
    ----------
    time : int
        Time (in seconds since simulation start) at which this event is executed.
    """
    def __init__(self, time: int, **kwds):
        super().__init__(start_time=time, end_time=time+1, **kwds)

    def get_attributes(self) -> dict:
        return {"time": self.start_time}


class PumpEvent(ActuatorEvent):
    """
    Base class of a pump event.

    Parameters
    ----------
    pump_id : str
        ID of the pump that is affected by this event.
    """
    def __init__(self, pump_id: str, **kwds):
        self.__pump_id = pump_id

        super().__init__(**kwds)

    def init(self, epanet_api: epanet) -> None:
        if self.__pump_id not in epanet_api.getLinkPumpNameID():
            raise ValueError(f"Invalid pump ID '{self.__pump_id}'")

        super().init(epanet_api)

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"pump_id": self.__pump_id}

    @property
    def pump_id(self) -> str:
        """
        Gets the ID of the pump affected by this event.

        Returns
        -------
        `str`
            Pump ID.
        """
        return self.__pump_id


@serializable(PUMP_STATE_EVENT_ID, ".epytflow_pump_state_event")
class PumpStateEvent(PumpEvent, JsonSerializable):
    """
    Class implementing a pump state event.

    Parameters
    ----------
    pump_state : `str`
        New state of the pump -- i.e. the state of the pump is set to this value
        while the event is active.

        Must be one of the following constants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED  = 0
            - EN_OPEN    = 1
    """
    def __init__(self, pump_state: int, **kwds):
        if not isinstance(pump_state, int):
            raise TypeError("'pump_state' must be an instace of 'int' " +
                            f"but not of {type(pump_state)}")
        if not 0 <= pump_state <= 1:
            raise ValueError(f"Invalid pump state '{pump_state}' -- " +
                             "must be either EN_CLOSED (0) or EN_OPEN (1)")

        self.__pump_state = pump_state

        super().__init__(**kwds)

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"pump_state": self.__pump_state}

    @property
    def pump_state(self) -> int:
        """
        Gets the new pump state.

        Returns
        -------
        `int`
            New pump state.

            One of the following constants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

                - EN_CLOSED  = 0
                - EN_OPEN    = 1
        """
        return self.__pump_state

    def apply(self, cur_time: int) -> None:
        pump_idx = self._epanet_api.getLinkPumpNameID().index(self.pump_id) + 1
        pump_link_idx = self._epanet_api.getLinkPumpIndex(pump_idx)

        pattern_idx = self._epanet_api.getLinkPumpPatternIndex(pump_idx)
        if pattern_idx != 0:
            warnings.warn(f"Can not set pump state of pump {self.pump_id} " +
                          "because a pump pattern exists")
        else:
            self._epanet_api.setLinkStatus(pump_link_idx, self.__pump_state)


@serializable(PUMP_SPEED_EVENT_ID, ".epytflow_pump_speed_event")
class PumpSpeedEvent(PumpEvent, JsonSerializable):
    """
    Class implementing a pump speed event.

    Parameters
    ----------
    pump_speed : float
        New pump speed -- i.e. the speed of the pump is set to this value while the event is active.
    """
    def __init__(self, pump_speed: float, **kwds):
        if not isinstance(pump_speed, float):
            raise TypeError("'pump_speed' must be an instance of 'float' " +
                            f"but not of {type(pump_speed)}")
        if pump_speed <= 0:
            raise ValueError("Pump speed must be positive")

        self.__pump_speed = pump_speed

        super().__init__(**kwds)

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"pump_speed": self.__pump_speed}

    @property
    def pump_speed(self) -> float:
        """
        Gets the new pump speed.

        Returns
        -------
        `float`
            New pump speed.
        """
        return self.__pump_speed

    def apply(self, cur_time: int) -> None:
        pump_idx = self._epanet_api.getLinkPumpNameID().index(self.pump_id) + 1
        pattern_idx = self._epanet_api.getLinkPumpPatternIndex(pump_idx)

        if pattern_idx == 0:
            warnings.warn(f"No pattern for pump '{self.pump_id}' found -- a new pattern is created")
            pattern_idx = self._epanet_api.addPattern(f"pump_speed_{self.pump_id}")
            self._epanet_api.setLinkPumpPatternIndex(pump_idx, pattern_idx)

        self._epanet_api.setPattern(pattern_idx, np.array([self.__pump_speed]))


@serializable(VALVE_STATE_EVENT_ID, ".epytflow_valve_state_event")
class ValveStateEvent(ActuatorEvent, JsonSerializable):
    """
    Class implementing a valve state event.

    Parameters
    ----------
    valve_id : `str`
        ID of the valve that is affected by this event.
    valve_state : `str`
        New state of the valve -- i.e. the valve state is set to this value this event is executed.

        Must be one of the following constants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED       = 0
            - EN_OPEN         = 1
    """
    def __init__(self, valve_id: str, valve_state: int, **kwds):
        if not isinstance(valve_state, int):
            raise TypeError("'valve_state' must be an instance of 'int' " +
                            f"but not of {type(valve_state)}")
        if not 0 <= valve_state <= 1:
            raise ValueError(f"Invalid valve state '{valve_state}' -- " +
                             "must be either EN_CLOSED (0) or EN_OPEN (1)")

        self.__valve_id = valve_id
        self.__valve_state = valve_state

        super().__init__(**kwds)

    def init(self, epanet_api: epanet) -> None:
        if self.__valve_id not in epanet_api.getLinkValveNameID():
            raise ValueError(f"Invalid valve ID '{self.__valve_id}'")

        super().init(epanet_api)

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"valve_id": self.__valve_id,
                                           "valve_state": self.__valve_state}

    @property
    def valve_id(self) -> str:
        """
        Gets the ID of the valve affected by this event.

        Returns
        -------
        `str`
            Valve ID.
        """
        return self.__valve_id

    @property
    def valve_state(self) -> int:
        """
        Gets the new state of the valve.

        Returns
        -------
        `int`
            New valve state.

            One of the following constants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:
        """
        return self.__valve_state

    def apply(self, cur_time: int) -> None:
        valve_idx = self._epanet_api.getLinkValveNameID().index(self.__valve_id) + 1
        valve_link_idx = self._epanet_api.getLinkValveIndex(valve_idx)
        self._epanet_api.setLinkStatus(valve_link_idx, self.__valve_state)



"""
The module contains classes for representing simple control rules as used in EPANET.
"""
from typing import Union
from epyt.epanet import ToolkitConstants

from ..events import ActuatorConstants
from ...serialization import JsonSerializable, SIMPLE_CONTROL_ID, serializable


@serializable(SIMPLE_CONTROL_ID, ".epytflow_simple_control")
class SimpleControlModule(JsonSerializable):
    """
    A class for representing a simple EPANET control rule.

    Parameters
    ----------
    link_id : `str`
        Link ID.
    link_status : `int` or `float`
        Status of the link that is set when the condition is fullfilled.

        Instance of `float` if the link constitutes a pump -- in this case,
        the argument corresponds to the pump speed.

        Instance of `int` if the link constitutes a valve -- in this case,
        must be one of the followig costants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED = 0
            - EN_OPEN   = 1
    cond_type : `int`
        Condition/Rule type.

        Must be one of the following EPANET constants:

            - EN_LOWLEVEL  = 0
            - EN_HILEVEL   = 1
            - EN_TIMER     = 2
            - EN_TIMEOFDAY = 3
    cond_var_value : `str` or `int`
        Condition/Rule variable value.

        Node ID in the cases of EN_LOWLEVEL or EN_HILEVEL.
        Time of the day (in AM/PM format) in the case of EN_TIMEOFDAY.
        Number of hours (as an integer) since simulation start in the case of EN_TIMER.
    cond_comp_value : `float`
        The condition/rule comparison value at which this control rule is triggered.

        Lower or upper value on the pressure (or tank level) in the cases of
        EN_LOWLEVEL and EN_HILEVEL.

        Will be ignored in all other cases -- i.e. should be set to None.
    """
    def __init__(self, link_id: str, link_status: Union[int, float], cond_type: int,
                 cond_var_value: Union[str, int], cond_comp_value: float,
                 **kwds):
        if not isinstance(link_id, str):
            raise TypeError(f"'link_id' must be an instance of 'str' but not of '{type(link_id)}'")
        if isinstance(link_status, int):
            if link_status not in [ActuatorConstants.EN_OPEN, ActuatorConstants.EN_CLOSED]:
                raise ValueError(f"Invalid link status {link_status} in 'link_status'")
        elif isinstance(link_status, float):
            if link_status < 0:
                raise TypeError("'link_status' can not be negative")
        else:
            raise TypeError("'link_status' must be an instance of 'int' or 'float' but not " +
                            f"of '{type(link_status)}'")
        if cond_type not in [ToolkitConstants.EN_TIMEOFDAY, ToolkitConstants.EN_TIMER,
                             ToolkitConstants.EN_LOWLEVEL, ToolkitConstants.EN_HILEVEL]:
            raise ValueError(f"Invalid control type '{cond_type}' in 'cond_type'")

        if cond_type == ToolkitConstants.EN_TIMEOFDAY:
            if not isinstance(cond_var_value, str):
                raise TypeError("EN_TIMEOFDAY requires that 'cond_var_value' must be an instance " +
                                f"of 'str' but not of '{type(cond_var_value)}'")
            if not cond_var_value.endswith("AM") and not cond_var_value.endswith("PM"):
                raise ValueError(f"Invalid time of day format '{cond_var_value}' in " +
                                 "'cond_var_value'")
        elif cond_type == ToolkitConstants.EN_TIMER:
            if not isinstance(cond_var_value, int):
                raise TypeError("EN_TIMER requires that 'cond_var_value' must be an instance " +
                                f"of 'int' but not of '{type(cond_var_value)}'")
            if cond_var_value < 0:
                raise ValueError("'cond_var_value' can not be negative")
        else:
            if not isinstance(cond_var_value, str):
                raise TypeError("'cond_var_value' must be an instance of 'str' but " +
                                f"not of '{type(cond_var_value)}'")
            if not isinstance(cond_comp_value, float):
                raise TypeError("'cond_comp_value' must be an instance of 'float' " +
                                f"but not of '{type(cond_comp_value)}'")
            if cond_comp_value < 0:
                raise ValueError("'cond_comp_value' can not be negative")

        self.__link_id = link_id
        self.__link_status = link_status
        self.__cond_type = cond_type
        self.__cond_var_value = cond_var_value
        self.__cond_comp_value = cond_comp_value

        super().__init__(**kwds)

    @property
    def link_id(self) -> str:
        """
        Returns the link ID.

        Returns
        -------
        `str`
            Link ID.
        """
        return self.__link_id

    @property
    def link_status(self) -> Union[int, float]:
        """
        Returns the link status that is set when the condition is fullfilled.

        Returns
        -------
        `int` or `float`
            Pump speed if the link is a pump, otherwise one of the followig costants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

                - EN_CLOSED = 0
                - EN_OPEN   = 1
        """
        return self.__link_status

    @property
    def cond_type(self) -> int:
        """
        Returns the condition/rule type.

        Returns
        -------
        `int`
            Condition/Rule type -- will be one of the following EPANET constants:

                - EN_LOWLEVEL  = 0
                - EN_HILEVEL   = 1
                - EN_TIMER     = 2
                - EN_TIMEOFDAY = 3
        """
        return self.__cond_type

    @property
    def cond_var_value(self) -> Union[str, int]:
        """
        Return the condition/rule variable value.

        Node ID in the cases of EN_LOWLEVEL or EN_HILEVEL.
        Time of the day (in AM/PM format) in the case of EN_TIMEOFDAY.
        Number of hours (as an integer) since simulation start in the case of EN_TIMER.

        Returns
        -------
        `str` or `int`
            Condition/rule variable value.
        """
        return self.__cond_var_value

    @property
    def cond_comp_value(self) -> float:
        """
        Returns the condition/rule comparison value -- might be None if not needed.

        Lower or upper value on the pressure (or tank level) in the cases of
        EN_LOWLEVEL and EN_HILEVEL.

        Returns
        -------
        `float`
            Condition/Rule comparison value.
        """
        return self.__cond_comp_value

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"link_id": self.__link_id,
                                           "link_status": self.__link_status,
                                           "cond_type": self.__cond_type,
                                           "cond_var_value": self.__cond_var_value,
                                           "cond_comp_value": self.__cond_comp_value}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.__link_id == other.link_id and \
            self.__link_status == other.link_status and self.__cond_type == other.cond_type and \
            self.__cond_var_value == other.cond_var_value and \
            self.__cond_comp_value == other.cond_comp_value

    def __str__(self) -> str:
        control_rule_str = f"LINK {self.__link_id} "
        if isinstance(self.__link_status, int):
            control_rule_str += "OPEN " if self.__link_status == ActuatorConstants.EN_OPEN \
                else "CLOSED "
        else:
            control_rule_str += f"{self.__link_status} "

        if self.__cond_type == ToolkitConstants.EN_TIMER:
            control_rule_str += f"AT TIME {self.__cond_var_value}"
        elif self.__cond_type == ToolkitConstants.EN_TIMEOFDAY:
            control_rule_str += f"AT CLOCKTIME {self.__cond_var_value}"
        elif self.__cond_type == ToolkitConstants.EN_LOWLEVEL:
            control_rule_str += f"IF NODE {self.__cond_var_value} BELOW {self.__cond_comp_value}"
        elif self.__cond_type == ToolkitConstants.EN_HILEVEL:
            control_rule_str += f"IF NODE {self.__cond_var_value} ABOVE {self.__cond_comp_value}"

        return control_rule_str


class SimplePumpSpeedTimeControl(SimpleControlModule):
    """
    A class for representing a simple control rule for setting the pump speed at some point in time.

    Parameters
    ----------
    pump_id : `str`
        Pump ID.
    pump_speed : `float`
        Pump speed.
    time : `str` or `int`
        Time of the day (in AM/PM format) in the case or
        number of hours (as an integer) since simulation start.
    """
    def __init__(self, pump_id: str, pump_speed: float, time: Union[str, int]):
        super().__init__(link_id=pump_id, link_status=pump_speed,
                         cond_type=ToolkitConstants.EN_TIMER if isinstance(time, int)
                         else ToolkitConstants.EN_TIMEOFDAY,
                         cond_var_value=time, cond_comp_value=None)


class SimplePumpSpeedConditionControl(SimpleControlModule):
    """
    A class for representing a simple IF-THEN control rule for setting the pump speed.

    Parameters
    ----------
    Parameters
    ----------
    pump_id : `str`
        Pump ID.
    pump_speed : `float`
        Pump speed.
    node_id : `str`
        Node ID.
    comp_type : `int`
        Comparison type -- must be one of the following EPANET constants:

            - EN_LOWLEVEL  = 0
            - EN_HILEVEL   = 1
    comp_value : `float`:
        Lower or upper value on the pressure (or tank level) at which this
        control rule is triggered.
    """
    def __init__(self, pump_id: str, pump_speed: float, node_id: str, comp_type: int,
                 comp_value: float):
        super().__init__(link_id=pump_id, link_status=pump_speed, cond_type=comp_type,
                         cond_var_value=node_id, cond_comp_value=comp_value)


class SimpleValveTimeControl(SimpleControlModule):
    """
    A class for representing a simple control rule for setting the valve status
    at some point in time.

    Parameters
    ----------
    valve_id : `str`
        valve ID.
    valve_status : `int`
        Valve status -- must be one of the followig costants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED = 0
            - EN_OPEN   = 1
    time : `str` or `int`
        Time of the day (in AM/PM format) in the case or
        number of hours (as an integer) since simulation start.
    """
    def __init__(self, valve_id: str, valve_status: int, time: Union[str, int]):
        super().__init__(link_id=valve_id, link_status=valve_status,
                         cond_type=ToolkitConstants.EN_TIMER if isinstance(time, int)
                         else ToolkitConstants.EN_TIMEOFDAY,
                         cond_var_value=time, cond_comp_value=None)


class SimpleValveConditionControl(SimpleControlModule):
    """
    A class for representing a simple IF-THEN control rule for setting the valve status.

    Parameters
    ----------
    valve_id : `str`
        valve ID.
    valve_status : `int`
        Valve status -- must be one of the followig costants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED = 0
            - EN_OPEN   = 1
    node_id : `str`
        Node ID.
    comp_type : `int`
        Comparison type -- must be one of the following EPANET constants:

            - EN_LOWLEVEL  = 0
            - EN_HILEVEL   = 1
    comp_value : `float`:
        Lower or upper value on the pressure (or tank level) at which this
        control rule is triggered.
    """
    def __init__(self, valve_id: str, valve_status: int, node_id: str, comp_type: int,
                 comp_value: float):
        super().__init__(link_id=valve_id, link_status=valve_status, cond_type=comp_type,
                         cond_var_value=node_id, cond_comp_value=comp_value)



"""
Module provides a base class for control environments.
"""
import os
import uuid
from abc import abstractmethod, ABC
from typing import Union
import warnings
import numpy as np

from ..simulation import ScenarioSimulator, ScenarioConfig, ScadaData, ToolkitConstants
from ..utils import get_temp_folder


class ScenarioControlEnv(ABC):
    """
    Base class for a control environment challenge.

    Parameters
    ----------
    scenario_config : :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Scenario configuration.
    autoreset : `bool`, optional
        If True, environment is automatically reset if terminated.

        The default is False.
    reapply_uncertainties_at_reset : `bool`, optional
        If True, the uncertainties are re-applied to the original properties at each reset.

        The default is False.

    Attributes
    ----------
    _scenario_sim : :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`, protected
        Scenario simulator of the control scenario.
    _scenario_config : :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Scenario configuration.
    _sim_generator : Generator[Union[:class:`~epyt_flow.simulation.scada.scada_data.ScadaData`, dict], bool, None], protected
        Generator for running the step-wise simulation.
    _hydraulic_scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`, protected
        SCADA data from the hydraulic simulation -- only used if EPANET-MSX is used in the control scenario.
    """
    def __init__(self, scenario_config: ScenarioConfig, autoreset: bool = False,
                 reapply_uncertainties_at_reset: bool = False,
                 **kwds):
        if not isinstance(scenario_config, ScenarioConfig):
            raise TypeError("'scenario_config' must be an instance of " +
                            "'epyt_flow.simulation.ScenarioConfig' " +
                            "but not of '{type(scenario_config)}'")
        if not isinstance(autoreset, bool):
            raise TypeError("'autoreset' must be an instance of 'bool' " +
                            f"but not of '{type(autoreset)}'")
        if not isinstance(reapply_uncertainties_at_reset, bool):
            raise TypeError("")

        self._scenario_config = scenario_config
        self._scenario_sim = None
        self._sim_generator = None
        self.__autoreset = autoreset
        self._hydraulic_scada_data = None
        self.__reapply_uncertainties_at_reset = reapply_uncertainties_at_reset

        super().__init__(**kwds)

    @property
    def autoreset(self) -> bool:
        """
        True, if environment automatically resets after it terminated.

        Returns
        -------
        `bool`
            True, if environment automatically resets after it terminated.
        """
        return self.__autoreset

    @property
    def reapply_uncertainties_at_reset(self) -> bool:
        """
        True, if the uncertainties are re-applied to the original properties at each reset.

        Returns
        -------
        `bool`
            True, if the uncertainties are re-applied to the original properties at each reset.
        """
        return self.__reapply_uncertainties_at_reset

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self) -> None:
        """
        Frees all resources.
        """
        try:
            if self._sim_generator is not None:
                next(self._sim_generator)
                self._sim_generator.send(True)
                self._sim_generator = None
        except StopIteration:
            pass

        if self._scenario_sim is not None:
            self._scenario_sim.close()
            self._scenario_sim = None

    def contains_events(self) -> bool:
        """
        Check if the scenario contains any events.

        Returns
        -------
        `bool`
            True is the scenario contains any events, False otherwise.
        """
        return len(self._scenario_config.system_events) != 0 or \
            len(self._scenario_config.sensor_reading_events) != 0

    def reset(self) -> Union[tuple[ScadaData, bool], ScadaData]:
        """
        Resets the environment (i.e. simulation).

        Returns
        -------
        :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            Current SCADA data (i.e. sensor readings).
        """
        self.close()

        self._scenario_sim = ScenarioSimulator(
            scenario_config=self._scenario_config)

        if self._scenario_sim.f_msx_in is not None:
            # Run hydraulic simulation first
            hyd_export = os.path.join(get_temp_folder(), f"epytflow_env_MSX_{uuid.uuid4()}.hyd")
            sim = self._scenario_sim.run_hydraulic_simulation
            self._hydraulic_scada_data = sim(hyd_export=hyd_export,
                                             reapply_uncertainties=self.__reapply_uncertainties_at_reset)

            # Run advanced quality analysis (EPANET-MSX) on top of the computed hydraulics
            gen = self._scenario_sim.run_advanced_quality_simulation_as_generator
            self._sim_generator = gen(hyd_export, support_abort=True)
        else:
            gen = self._scenario_sim.run_hydraulic_simulation_as_generator
            self._sim_generator = gen(support_abort=True,
                                      reapply_uncertainties=self.__reapply_uncertainties_at_reset)

        return self._next_sim_itr()

    def _next_sim_itr(self) -> Union[tuple[ScadaData, bool], ScadaData]:
        try:
            next(self._sim_generator)
            scada_data, terminated = self._sim_generator.send(False)

            if self._scenario_sim.f_msx_in is not None:
                cur_time = int(scada_data.sensor_readings_time[0])
                cur_hyd_scada_data = self._hydraulic_scada_data.\
                    extract_time_window(cur_time, cur_time)
                scada_data.join(cur_hyd_scada_data)

            if self.autoreset is True:
                return scada_data
            else:
                return scada_data, terminated
        except StopIteration:
            if self.__autoreset is True:
                return self.reset()
            else:
                return None, True

    def set_pump_status(self, pump_id: str, status: int) -> None:
        """
        Sets the status of a pump.

        Parameters
        ----------
        pump_id : `str`
            ID of the pump for which the status is set.
        status : `int`
            New status of the pump -- either active (i.e. open) or inactive (i.e. closed).

            Must be one of the following constants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

                - EN_CLOSED  = 0
                - EN_OPEN    = 1
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        pump_idx = self._scenario_sim.epanet_api.getLinkPumpNameID().index(pump_id) + 1
        pump_link_idx = self._scenario_sim.epanet_api.getLinkPumpIndex(pump_idx)

        pattern_idx = self._scenario_sim.epanet_api.getLinkPumpPatternIndex(pump_idx)
        if pattern_idx != 0:
            warnings.warn(f"Can not set pump state of pump {pump_id} because a pump pattern exists")
        else:
            self._scenario_sim.epanet_api.setLinkStatus(pump_link_idx, status)

    def set_pump_speed(self, pump_id: str, speed: float) -> None:
        """
        Sets the speed of a pump.

        Parameters
        ----------
        pump_id : `str`
            ID of the pump for which the pump speed is set.
        speed : `float`
            New pump speed.
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        pump_idx = self._scenario_sim.epanet_api.getLinkPumpNameID().index(pump_id)
        pattern_idx = self._scenario_sim.epanet_api.getLinkPumpPatternIndex(pump_idx + 1)

        if pattern_idx == 0:
            warnings.warn(f"No pattern for pump '{pump_id}' found -- a new pattern is created")
            pattern_idx = self._scenario_sim.epanet_api.addPattern(f"pump_speed_{pump_id}")
            self._scenario_sim.epanet_api.setLinkPumpPatternIndex(pump_idx + 1, pattern_idx)

        self._scenario_sim.epanet_api.setPattern(pattern_idx, np.array([speed]))

    def set_valve_status(self, valve_id: str, status: int) -> None:
        """
        Sets the status of a valve.

        Parameters
        ----------
        valve_id : `str`
            ID of the valve for which the status is set.
        status : `int`
            New status of the valve -- either open or closed.

            Must be one of the following constants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

                - EN_CLOSED  = 0
                - EN_OPEN    = 1
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        valve_idx = self._scenario_sim.epanet_api.getLinkValveNameID().index(valve_id)
        valve_link_idx = self._scenario_sim.epanet_api.getLinkValveIndex()[valve_idx]
        self._scenario_sim.epanet_api.setLinkStatus(valve_link_idx, status)

    def set_node_quality_source_value(self, node_id: str, pattern_id: str,
                                      qual_value: float) -> None:
        """
        Sets the quality source at a particular node to a specific value -- e.g.
        setting the chlorine concentration injection to a specified value.

        Parameters
        ----------
        node_id : `str`
            ID of the node.
        pattern_id : `str`
            ID of the quality pattern at the specific node.
        qual_value : `float`
            New quality source value.
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        node_idx = self._scenario_sim.epanet_api.getNodeIndex(node_id)
        pattern_idx = self._scenario_sim.epanet_api.getPatternIndex(pattern_id)
        self._scenario_sim.epanet_api.setNodeSourceQuality(node_idx, 1)
        self._scenario_sim.epanet_api.setPattern(pattern_idx, np.array([qual_value]))

    def set_node_species_source_value(self, species_id: str, node_id: str, source_type: int,
                                      pattern_id: str, source_strength: float) -> None:
        """
        Sets the species source at a particular node to a specific value -- i.e.
        setting the species injection amount at a particular location.

        Parameters
        ----------
        species_id : `str`
            ID of the species.
        node_id : `str`
            ID of the node.
        source_type : `int`
            Type of the external species injection source -- must be one of
            the following EPANET toolkit constants:

                - EN_CONCEN     = 0
                - EN_MASS       = 1
                - EN_SETPOINT   = 2
                - EN_FLOWPACED  = 3

            Description:

                - E_CONCEN Sets the concentration of external inflow entering a node
                - EN_MASS Injects a given mass/minute into a node
                - EN_SETPOINT Sets the concentration leaving a node to a given value
                - EN_FLOWPACED Adds a given value to the concentration leaving a node
        pattern_id : `str`
            ID of the source pattern.
        source_strength : `float`
            Amount of the injected species (source strength) --
            i.e. interpreation of this number depends on `source_type`
        """
        if self._scenario_sim.f_msx_in is None:
            raise RuntimeError("You are not running EPANET-MSX")

        source_type_ = "None"
        if source_type == ToolkitConstants.EN_CONCEN:
            source_type_ = "CONCEN"
        elif source_type == ToolkitConstants.EN_MASS:
            source_type_ = "MASS"
        elif source_type == ToolkitConstants.EN_SETPOINT:
            source_type_ = "SETPOINT"
        elif source_type == ToolkitConstants.EN_FLOWPACED:
            source_type_ = "FLOWPACED"

        self._scenario_sim.epanet_api.setMSXPattern(pattern_id, [1])
        self._scenario_sim.epanet_api.setMSXSources(node_id, species_id, source_type_,
                                                    source_strength, pattern_id)

    @abstractmethod
    def step(self, *actions) -> Union[tuple[ScadaData, float, bool], tuple[ScadaData, float]]:
        """
        Performs the next step by applying an action and observing
        the consequences (SCADA data, reward, terminated).

        Note that `terminated` is only returned if `autoreset=False` otherwise
        only the current SCADA data and reward are returned.

        Returns
        -------
        `(` :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` `, float, bool)` or `(` :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` `, float)`
            Triple or tuple of observations (:class:`~epyt_flow.simulation.scada.scada_data.ScadaData`),
            reward (`float`), and terminated (`bool`).
        """
        raise NotImplementedError()



"""
Module provides a class for implementing sensor noise (e.g. uncertainty in sensor readings).
"""
from typing import Optional, Callable
from copy import deepcopy
import numpy
from numpy.random import default_rng

from .uncertainties import Uncertainty
from ..serialization import serializable, JsonSerializable, SENSOR_NOISE_ID


@serializable(SENSOR_NOISE_ID, ".epytflow_sensor_noise")
class SensorNoise(JsonSerializable):
    """
    Class implementing sensor noise/uncertainty.

    Parameters
    ----------
    global_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global sensor uncertainty. If None, no global sensor uncertainties are applied.

        The default is None.
    local_uncertainties : dict[tuple[int, str], :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local (i.e. sensor specific) uncertainties.
        If None, no local sensor uncertainties are applied.

        The default is None.
    seed : `int`, optional
        Seed for the random number generator.

        The default is None.
    """
    def __init__(self, global_uncertainty: Optional[Uncertainty] = None,
                 local_uncertainties: Optional[dict[int, str, Uncertainty]] = None,
                 seed: Optional[int] = None,
                 **kwds):
        if not isinstance(global_uncertainty, Uncertainty):
            raise TypeError("'global_uncertainty' must be an instance of " +
                            "'epyt_flow.uncertainty.Uncertainty' but not of " +
                            f"'{type(global_uncertainty)}'")
        if local_uncertainties is not None:
            if not isinstance(local_uncertainties, dict):
                raise TypeError("'local_uncertainties' must be an instance of " +
                                "'dict[tuple[int, str], epyt_flow.uncertainty.Uncertainty]' "+
                                f"but not of '{type(local_uncertainties)}'")
            if any(not isinstance(key[0], int) or not isinstance(key[1], str) or
                   not isinstance(local_uncertainties[key], Uncertainty)
                   for key in local_uncertainties.keys()):
                raise TypeError("'local_uncertainties' must be an instance of " +
                                "'dict[tuple[int, str], epyt_flow.uncertainty.Uncertainty]'")

        self.__global_uncertainty = global_uncertainty
        self.__local_uncertainties = local_uncertainties
        self.__np_rand_gen = default_rng(seed)

        super().__init__(**kwds)

    @property
    def global_uncertainty(self) -> Uncertainty:
        """
        Returns the global sensor readings uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global sensor readings uncertainty.
        """
        return deepcopy(self.__global_uncertainty)

    @property
    def local_uncertainties(self) -> dict[int, str, Uncertainty]:
        """
        Returns the local (i.e. sensor specific) uncertainties.

        Returns
        -------
        dict[tuple[int, str], :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local (i.e. sensor specific) uncertainties.
        """
        return deepcopy(self.__local_uncertainties)

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"global_uncertainty": self.__global_uncertainty,
                                           "local_uncertainties": self.__local_uncertainties}

    def __eq__(self, other) -> bool:
        if not isinstance(other, SensorNoise):
            raise TypeError("Can not compare 'SensorNoise' instance " +
                            f"with '{type(other)}' instance")

        return super().__eq__(other) and self.__global_uncertainty == other.global_uncertainty and \
            self.__local_uncertainties == other.local_uncertainties

    def __str__(self) -> str:
        return f"global_uncertainty: {self.__global_uncertainty} " + \
            f"local_uncertainties: {self.__local_uncertainties}"

    def apply_local_uncertainty(self, map_sensor_to_idx: Callable[[int, str], int],
                                sensor_readings: numpy.ndarray) -> numpy.ndarray:
        """
        Applies the local (i.e. sensor specific) sensor uncertainties -- i.e. sensor readings
        are perturbed according to the specified uncertainties.

        Parameters
        ----------
        map_sensor_to_idx : `Callable[[int, str], int]`
            Function mapping sensor type (int) and sensor id (e.g. node id, link id, etc.) to indices
            in the final sensor readings.
        sensor_readings : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            All (global) sensor readings (no matter if ther).

        Returns
        -------
        `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Perturbed sensor readings.
        """
        if self.__local_uncertainties is None:
            return sensor_readings
        else:
            self.__local_uncertainties.set_random_generator(self.__np_rand_gen)

            for (sensor_type, sensor_id), uncertainty in map_sensor_to_idx.items():
                idx = map_sensor_to_idx(sensor_type, sensor_id)
                sensor_readings[:, idx] = uncertainty.apply_batch(sensor_readings[:, idx])
                return sensor_readings

    def apply_global_uncertainty(self, sensor_readings: numpy.ndarray) -> numpy.ndarray:
        """
        Applies the global sensor uncertainty to given sensor readings -- i.e. sensor readings
        are perturbed according to the specified uncertainty.

        .. note::
            Note that state sensor readings such as valve states, pump states, etc.
            are NOT affected by sensor noise!

        Parameters
        ----------
        sensor_readings : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            All (global) senor readings.

        Returns
        -------
        `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Perturbed sensor readings.
        """
        if self.__global_uncertainty is None:
            return sensor_readings
        else:
            self.__global_uncertainty.set_random_generator(self.__np_rand_gen)

            return self.__global_uncertainty.apply_batch(sensor_readings)
